"""A web3.py Contract class for the IMorpho contract.

DO NOT EDIT.  This file was generated by pypechain.  See documentation at
https://github.com/delvtech/pypechain"""

# contracts have PascalCase names
# pylint: disable=invalid-name

# contracts control how many attributes and arguments we have in generated code
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-arguments

# we don't need else statement if the other conditionals all have return,
# but it's easier to generate
# pylint: disable=no-else-return

# This file is bound to get very long depending on contract sizes.
# pylint: disable=too-many-lines

# methods are overriden with specific arguments instead of generic *args, **kwargs
# pylint: disable=arguments-differ

# consumers have too many opinions on line length
# pylint: disable=line-too-long


from __future__ import annotations

from typing import Any, NamedTuple, Type, cast

from eth_account.signers.local import LocalAccount
from eth_typing import ChecksumAddress, HexStr
from typing_extensions import Self
from web3 import Web3
from web3.contract.contract import (
    Contract,
    ContractConstructor,
    ContractFunction,
    ContractFunctions,
)
from web3.exceptions import FallbackNotFound
from web3.types import ABI, BlockIdentifier, CallOverride, TxParams

from .AuthorizationTypes import Authorization
from .MarketParamsTypes import MarketParams
from .MarketTypes import Market
from .PositionTypes import Position
from .SignatureTypes import Signature
from .utilities import (
    dataclass_to_tuple,
    rename_returned_types,
    try_bytecode_hexbytes,
)

structs = {
    "MarketParams": MarketParams,
    "Market": Market,
    "Position": Position,
    "Authorization": Authorization,
    "Signature": Signature,
}


class IMorphoDOMAIN_SEPARATORContractFunction(ContractFunction):
    """ContractFunction for the DOMAIN_SEPARATOR method."""

    def __call__(self) -> IMorphoDOMAIN_SEPARATORContractFunction:  # type: ignore
        clone = super().__call__()
        self.kwargs = clone.kwargs
        self.args = clone.args
        return self

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> bytes:
        """returns bytes."""
        # Define the expected return types from the smart contract call

        return_types = bytes

        # Call the function

        raw_values = super().call(
            transaction, block_identifier, state_override, ccip_read_enabled
        )
        return cast(
            bytes, rename_returned_types(structs, return_types, raw_values)
        )


class IMorphoAccrueInterestContractFunction(ContractFunction):
    """ContractFunction for the accrueInterest method."""

    def __call__(self, marketParams: MarketParams) -> IMorphoAccrueInterestContractFunction:  # type: ignore
        clone = super().__call__(dataclass_to_tuple(marketParams))
        self.kwargs = clone.kwargs
        self.args = clone.args
        return self

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> None:
        """returns None."""
        # Define the expected return types from the smart contract call

        # Call the function


class IMorphoBorrowContractFunction(ContractFunction):
    """ContractFunction for the borrow method."""

    class ReturnValues(NamedTuple):
        """The return named tuple for Borrow."""

        assetsBorrowed: int
        sharesBorrowed: int

    def __call__(self, marketParams: MarketParams, assets: int, shares: int, onBehalf: str, receiver: str) -> IMorphoBorrowContractFunction:  # type: ignore
        clone = super().__call__(
            dataclass_to_tuple(marketParams),
            dataclass_to_tuple(assets),
            dataclass_to_tuple(shares),
            dataclass_to_tuple(onBehalf),
            dataclass_to_tuple(receiver),
        )
        self.kwargs = clone.kwargs
        self.args = clone.args
        return self

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> ReturnValues:
        """returns ReturnValues."""
        # Define the expected return types from the smart contract call

        return_types = [int, int]

        # Call the function

        raw_values = super().call(
            transaction, block_identifier, state_override, ccip_read_enabled
        )
        return self.ReturnValues(
            *rename_returned_types(structs, return_types, raw_values)
        )


class IMorphoCreateMarketContractFunction(ContractFunction):
    """ContractFunction for the createMarket method."""

    def __call__(self, marketParams: MarketParams) -> IMorphoCreateMarketContractFunction:  # type: ignore
        clone = super().__call__(dataclass_to_tuple(marketParams))
        self.kwargs = clone.kwargs
        self.args = clone.args
        return self

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> None:
        """returns None."""
        # Define the expected return types from the smart contract call

        # Call the function


class IMorphoEnableIrmContractFunction(ContractFunction):
    """ContractFunction for the enableIrm method."""

    def __call__(self, irm: str) -> IMorphoEnableIrmContractFunction:  # type: ignore
        clone = super().__call__(dataclass_to_tuple(irm))
        self.kwargs = clone.kwargs
        self.args = clone.args
        return self

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> None:
        """returns None."""
        # Define the expected return types from the smart contract call

        # Call the function


class IMorphoEnableLltvContractFunction(ContractFunction):
    """ContractFunction for the enableLltv method."""

    def __call__(self, lltv: int) -> IMorphoEnableLltvContractFunction:  # type: ignore
        clone = super().__call__(dataclass_to_tuple(lltv))
        self.kwargs = clone.kwargs
        self.args = clone.args
        return self

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> None:
        """returns None."""
        # Define the expected return types from the smart contract call

        # Call the function


class IMorphoExtSloadsContractFunction(ContractFunction):
    """ContractFunction for the extSloads method."""

    def __call__(self, slots: list[bytes]) -> IMorphoExtSloadsContractFunction:  # type: ignore
        clone = super().__call__(dataclass_to_tuple(slots))
        self.kwargs = clone.kwargs
        self.args = clone.args
        return self

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> list[bytes]:
        """returns list[bytes]."""
        # Define the expected return types from the smart contract call

        return_types = list[bytes]

        # Call the function

        raw_values = super().call(
            transaction, block_identifier, state_override, ccip_read_enabled
        )
        return cast(
            list[bytes],
            rename_returned_types(structs, return_types, raw_values),
        )


class IMorphoFeeRecipientContractFunction(ContractFunction):
    """ContractFunction for the feeRecipient method."""

    def __call__(self) -> IMorphoFeeRecipientContractFunction:  # type: ignore
        clone = super().__call__()
        self.kwargs = clone.kwargs
        self.args = clone.args
        return self

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> str:
        """returns str."""
        # Define the expected return types from the smart contract call

        return_types = str

        # Call the function

        raw_values = super().call(
            transaction, block_identifier, state_override, ccip_read_enabled
        )
        return cast(
            str, rename_returned_types(structs, return_types, raw_values)
        )


class IMorphoFlashLoanContractFunction(ContractFunction):
    """ContractFunction for the flashLoan method."""

    def __call__(self, token: str, assets: int, data: bytes) -> IMorphoFlashLoanContractFunction:  # type: ignore
        clone = super().__call__(
            dataclass_to_tuple(token),
            dataclass_to_tuple(assets),
            dataclass_to_tuple(data),
        )
        self.kwargs = clone.kwargs
        self.args = clone.args
        return self

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> None:
        """returns None."""
        # Define the expected return types from the smart contract call

        # Call the function


class IMorphoIdToMarketParamsContractFunction(ContractFunction):
    """ContractFunction for the idToMarketParams method."""

    def __call__(self, _id: bytes) -> IMorphoIdToMarketParamsContractFunction:  # type: ignore
        clone = super().__call__(dataclass_to_tuple(_id))
        self.kwargs = clone.kwargs
        self.args = clone.args
        return self

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> MarketParams:
        """returns MarketParams."""
        # Define the expected return types from the smart contract call

        return_types = MarketParams

        # Call the function

        raw_values = super().call(
            transaction, block_identifier, state_override, ccip_read_enabled
        )
        return cast(
            MarketParams,
            rename_returned_types(structs, return_types, raw_values),
        )


class IMorphoIsAuthorizedContractFunction(ContractFunction):
    """ContractFunction for the isAuthorized method."""

    def __call__(self, authorizer: str, authorized: str) -> IMorphoIsAuthorizedContractFunction:  # type: ignore
        clone = super().__call__(
            dataclass_to_tuple(authorizer), dataclass_to_tuple(authorized)
        )
        self.kwargs = clone.kwargs
        self.args = clone.args
        return self

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> bool:
        """returns bool."""
        # Define the expected return types from the smart contract call

        return_types = bool

        # Call the function

        raw_values = super().call(
            transaction, block_identifier, state_override, ccip_read_enabled
        )
        return cast(
            bool, rename_returned_types(structs, return_types, raw_values)
        )


class IMorphoIsIrmEnabledContractFunction(ContractFunction):
    """ContractFunction for the isIrmEnabled method."""

    def __call__(self, irm: str) -> IMorphoIsIrmEnabledContractFunction:  # type: ignore
        clone = super().__call__(dataclass_to_tuple(irm))
        self.kwargs = clone.kwargs
        self.args = clone.args
        return self

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> bool:
        """returns bool."""
        # Define the expected return types from the smart contract call

        return_types = bool

        # Call the function

        raw_values = super().call(
            transaction, block_identifier, state_override, ccip_read_enabled
        )
        return cast(
            bool, rename_returned_types(structs, return_types, raw_values)
        )


class IMorphoIsLltvEnabledContractFunction(ContractFunction):
    """ContractFunction for the isLltvEnabled method."""

    def __call__(self, lltv: int) -> IMorphoIsLltvEnabledContractFunction:  # type: ignore
        clone = super().__call__(dataclass_to_tuple(lltv))
        self.kwargs = clone.kwargs
        self.args = clone.args
        return self

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> bool:
        """returns bool."""
        # Define the expected return types from the smart contract call

        return_types = bool

        # Call the function

        raw_values = super().call(
            transaction, block_identifier, state_override, ccip_read_enabled
        )
        return cast(
            bool, rename_returned_types(structs, return_types, raw_values)
        )


class IMorphoLiquidateContractFunction(ContractFunction):
    """ContractFunction for the liquidate method."""

    class ReturnValues(NamedTuple):
        """The return named tuple for Liquidate."""

        arg1: int
        arg2: int

    def __call__(self, marketParams: MarketParams, borrower: str, seizedAssets: int, repaidShares: int, data: bytes) -> IMorphoLiquidateContractFunction:  # type: ignore
        clone = super().__call__(
            dataclass_to_tuple(marketParams),
            dataclass_to_tuple(borrower),
            dataclass_to_tuple(seizedAssets),
            dataclass_to_tuple(repaidShares),
            dataclass_to_tuple(data),
        )
        self.kwargs = clone.kwargs
        self.args = clone.args
        return self

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> ReturnValues:
        """returns ReturnValues."""
        # Define the expected return types from the smart contract call

        return_types = [int, int]

        # Call the function

        raw_values = super().call(
            transaction, block_identifier, state_override, ccip_read_enabled
        )
        return self.ReturnValues(
            *rename_returned_types(structs, return_types, raw_values)
        )


class IMorphoMarketContractFunction(ContractFunction):
    """ContractFunction for the market method."""

    def __call__(self, _id: bytes) -> IMorphoMarketContractFunction:  # type: ignore
        clone = super().__call__(dataclass_to_tuple(_id))
        self.kwargs = clone.kwargs
        self.args = clone.args
        return self

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> Market:
        """returns Market."""
        # Define the expected return types from the smart contract call

        return_types = Market

        # Call the function

        raw_values = super().call(
            transaction, block_identifier, state_override, ccip_read_enabled
        )
        return cast(
            Market, rename_returned_types(structs, return_types, raw_values)
        )


class IMorphoNonceContractFunction(ContractFunction):
    """ContractFunction for the nonce method."""

    def __call__(self, authorizer: str) -> IMorphoNonceContractFunction:  # type: ignore
        clone = super().__call__(dataclass_to_tuple(authorizer))
        self.kwargs = clone.kwargs
        self.args = clone.args
        return self

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> int:
        """returns int."""
        # Define the expected return types from the smart contract call

        return_types = int

        # Call the function

        raw_values = super().call(
            transaction, block_identifier, state_override, ccip_read_enabled
        )
        return cast(
            int, rename_returned_types(structs, return_types, raw_values)
        )


class IMorphoOwnerContractFunction(ContractFunction):
    """ContractFunction for the owner method."""

    def __call__(self) -> IMorphoOwnerContractFunction:  # type: ignore
        clone = super().__call__()
        self.kwargs = clone.kwargs
        self.args = clone.args
        return self

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> str:
        """returns str."""
        # Define the expected return types from the smart contract call

        return_types = str

        # Call the function

        raw_values = super().call(
            transaction, block_identifier, state_override, ccip_read_enabled
        )
        return cast(
            str, rename_returned_types(structs, return_types, raw_values)
        )


class IMorphoPositionContractFunction(ContractFunction):
    """ContractFunction for the position method."""

    def __call__(self, _id: bytes, user: str) -> IMorphoPositionContractFunction:  # type: ignore
        clone = super().__call__(
            dataclass_to_tuple(_id), dataclass_to_tuple(user)
        )
        self.kwargs = clone.kwargs
        self.args = clone.args
        return self

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> Position:
        """returns Position."""
        # Define the expected return types from the smart contract call

        return_types = Position

        # Call the function

        raw_values = super().call(
            transaction, block_identifier, state_override, ccip_read_enabled
        )
        return cast(
            Position, rename_returned_types(structs, return_types, raw_values)
        )


class IMorphoRepayContractFunction(ContractFunction):
    """ContractFunction for the repay method."""

    class ReturnValues(NamedTuple):
        """The return named tuple for Repay."""

        assetsRepaid: int
        sharesRepaid: int

    def __call__(self, marketParams: MarketParams, assets: int, shares: int, onBehalf: str, data: bytes) -> IMorphoRepayContractFunction:  # type: ignore
        clone = super().__call__(
            dataclass_to_tuple(marketParams),
            dataclass_to_tuple(assets),
            dataclass_to_tuple(shares),
            dataclass_to_tuple(onBehalf),
            dataclass_to_tuple(data),
        )
        self.kwargs = clone.kwargs
        self.args = clone.args
        return self

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> ReturnValues:
        """returns ReturnValues."""
        # Define the expected return types from the smart contract call

        return_types = [int, int]

        # Call the function

        raw_values = super().call(
            transaction, block_identifier, state_override, ccip_read_enabled
        )
        return self.ReturnValues(
            *rename_returned_types(structs, return_types, raw_values)
        )


class IMorphoSetAuthorizationContractFunction(ContractFunction):
    """ContractFunction for the setAuthorization method."""

    def __call__(self, authorized: str, newIsAuthorized: bool) -> IMorphoSetAuthorizationContractFunction:  # type: ignore
        clone = super().__call__(
            dataclass_to_tuple(authorized), dataclass_to_tuple(newIsAuthorized)
        )
        self.kwargs = clone.kwargs
        self.args = clone.args
        return self

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> None:
        """returns None."""
        # Define the expected return types from the smart contract call

        # Call the function


class IMorphoSetAuthorizationWithSigContractFunction(ContractFunction):
    """ContractFunction for the setAuthorizationWithSig method."""

    def __call__(self, authorization: Authorization, signature: Signature) -> IMorphoSetAuthorizationWithSigContractFunction:  # type: ignore
        clone = super().__call__(
            dataclass_to_tuple(authorization), dataclass_to_tuple(signature)
        )
        self.kwargs = clone.kwargs
        self.args = clone.args
        return self

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> None:
        """returns None."""
        # Define the expected return types from the smart contract call

        # Call the function


class IMorphoSetFeeContractFunction(ContractFunction):
    """ContractFunction for the setFee method."""

    def __call__(self, marketParams: MarketParams, newFee: int) -> IMorphoSetFeeContractFunction:  # type: ignore
        clone = super().__call__(
            dataclass_to_tuple(marketParams), dataclass_to_tuple(newFee)
        )
        self.kwargs = clone.kwargs
        self.args = clone.args
        return self

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> None:
        """returns None."""
        # Define the expected return types from the smart contract call

        # Call the function


class IMorphoSetFeeRecipientContractFunction(ContractFunction):
    """ContractFunction for the setFeeRecipient method."""

    def __call__(self, newFeeRecipient: str) -> IMorphoSetFeeRecipientContractFunction:  # type: ignore
        clone = super().__call__(dataclass_to_tuple(newFeeRecipient))
        self.kwargs = clone.kwargs
        self.args = clone.args
        return self

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> None:
        """returns None."""
        # Define the expected return types from the smart contract call

        # Call the function


class IMorphoSetOwnerContractFunction(ContractFunction):
    """ContractFunction for the setOwner method."""

    def __call__(self, newOwner: str) -> IMorphoSetOwnerContractFunction:  # type: ignore
        clone = super().__call__(dataclass_to_tuple(newOwner))
        self.kwargs = clone.kwargs
        self.args = clone.args
        return self

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> None:
        """returns None."""
        # Define the expected return types from the smart contract call

        # Call the function


class IMorphoSupplyContractFunction(ContractFunction):
    """ContractFunction for the supply method."""

    class ReturnValues(NamedTuple):
        """The return named tuple for Supply."""

        assetsSupplied: int
        sharesSupplied: int

    def __call__(self, marketParams: MarketParams, assets: int, shares: int, onBehalf: str, data: bytes) -> IMorphoSupplyContractFunction:  # type: ignore
        clone = super().__call__(
            dataclass_to_tuple(marketParams),
            dataclass_to_tuple(assets),
            dataclass_to_tuple(shares),
            dataclass_to_tuple(onBehalf),
            dataclass_to_tuple(data),
        )
        self.kwargs = clone.kwargs
        self.args = clone.args
        return self

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> ReturnValues:
        """returns ReturnValues."""
        # Define the expected return types from the smart contract call

        return_types = [int, int]

        # Call the function

        raw_values = super().call(
            transaction, block_identifier, state_override, ccip_read_enabled
        )
        return self.ReturnValues(
            *rename_returned_types(structs, return_types, raw_values)
        )


class IMorphoSupplyCollateralContractFunction(ContractFunction):
    """ContractFunction for the supplyCollateral method."""

    def __call__(self, marketParams: MarketParams, assets: int, onBehalf: str, data: bytes) -> IMorphoSupplyCollateralContractFunction:  # type: ignore
        clone = super().__call__(
            dataclass_to_tuple(marketParams),
            dataclass_to_tuple(assets),
            dataclass_to_tuple(onBehalf),
            dataclass_to_tuple(data),
        )
        self.kwargs = clone.kwargs
        self.args = clone.args
        return self

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> None:
        """returns None."""
        # Define the expected return types from the smart contract call

        # Call the function


class IMorphoWithdrawContractFunction(ContractFunction):
    """ContractFunction for the withdraw method."""

    class ReturnValues(NamedTuple):
        """The return named tuple for Withdraw."""

        assetsWithdrawn: int
        sharesWithdrawn: int

    def __call__(self, marketParams: MarketParams, assets: int, shares: int, onBehalf: str, receiver: str) -> IMorphoWithdrawContractFunction:  # type: ignore
        clone = super().__call__(
            dataclass_to_tuple(marketParams),
            dataclass_to_tuple(assets),
            dataclass_to_tuple(shares),
            dataclass_to_tuple(onBehalf),
            dataclass_to_tuple(receiver),
        )
        self.kwargs = clone.kwargs
        self.args = clone.args
        return self

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> ReturnValues:
        """returns ReturnValues."""
        # Define the expected return types from the smart contract call

        return_types = [int, int]

        # Call the function

        raw_values = super().call(
            transaction, block_identifier, state_override, ccip_read_enabled
        )
        return self.ReturnValues(
            *rename_returned_types(structs, return_types, raw_values)
        )


class IMorphoWithdrawCollateralContractFunction(ContractFunction):
    """ContractFunction for the withdrawCollateral method."""

    def __call__(self, marketParams: MarketParams, assets: int, onBehalf: str, receiver: str) -> IMorphoWithdrawCollateralContractFunction:  # type: ignore
        clone = super().__call__(
            dataclass_to_tuple(marketParams),
            dataclass_to_tuple(assets),
            dataclass_to_tuple(onBehalf),
            dataclass_to_tuple(receiver),
        )
        self.kwargs = clone.kwargs
        self.args = clone.args
        return self

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier = "latest",
        state_override: CallOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> None:
        """returns None."""
        # Define the expected return types from the smart contract call

        # Call the function


class IMorphoContractFunctions(ContractFunctions):
    """ContractFunctions for the IMorpho contract."""

    DOMAIN_SEPARATOR: IMorphoDOMAIN_SEPARATORContractFunction

    accrueInterest: IMorphoAccrueInterestContractFunction

    borrow: IMorphoBorrowContractFunction

    createMarket: IMorphoCreateMarketContractFunction

    enableIrm: IMorphoEnableIrmContractFunction

    enableLltv: IMorphoEnableLltvContractFunction

    extSloads: IMorphoExtSloadsContractFunction

    feeRecipient: IMorphoFeeRecipientContractFunction

    flashLoan: IMorphoFlashLoanContractFunction

    idToMarketParams: IMorphoIdToMarketParamsContractFunction

    isAuthorized: IMorphoIsAuthorizedContractFunction

    isIrmEnabled: IMorphoIsIrmEnabledContractFunction

    isLltvEnabled: IMorphoIsLltvEnabledContractFunction

    liquidate: IMorphoLiquidateContractFunction

    market: IMorphoMarketContractFunction

    nonce: IMorphoNonceContractFunction

    owner: IMorphoOwnerContractFunction

    position: IMorphoPositionContractFunction

    repay: IMorphoRepayContractFunction

    setAuthorization: IMorphoSetAuthorizationContractFunction

    setAuthorizationWithSig: IMorphoSetAuthorizationWithSigContractFunction

    setFee: IMorphoSetFeeContractFunction

    setFeeRecipient: IMorphoSetFeeRecipientContractFunction

    setOwner: IMorphoSetOwnerContractFunction

    supply: IMorphoSupplyContractFunction

    supplyCollateral: IMorphoSupplyCollateralContractFunction

    withdraw: IMorphoWithdrawContractFunction

    withdrawCollateral: IMorphoWithdrawCollateralContractFunction

    def __init__(
        self,
        abi: ABI,
        w3: "Web3",
        address: ChecksumAddress | None = None,
        decode_tuples: bool | None = False,
    ) -> None:
        super().__init__(abi, w3, address, decode_tuples)
        self.DOMAIN_SEPARATOR = IMorphoDOMAIN_SEPARATORContractFunction.factory(
            "DOMAIN_SEPARATOR",
            w3=w3,
            contract_abi=abi,
            address=address,
            decode_tuples=decode_tuples,
            function_identifier="DOMAIN_SEPARATOR",
        )
        self.accrueInterest = IMorphoAccrueInterestContractFunction.factory(
            "accrueInterest",
            w3=w3,
            contract_abi=abi,
            address=address,
            decode_tuples=decode_tuples,
            function_identifier="accrueInterest",
        )
        self.borrow = IMorphoBorrowContractFunction.factory(
            "borrow",
            w3=w3,
            contract_abi=abi,
            address=address,
            decode_tuples=decode_tuples,
            function_identifier="borrow",
        )
        self.createMarket = IMorphoCreateMarketContractFunction.factory(
            "createMarket",
            w3=w3,
            contract_abi=abi,
            address=address,
            decode_tuples=decode_tuples,
            function_identifier="createMarket",
        )
        self.enableIrm = IMorphoEnableIrmContractFunction.factory(
            "enableIrm",
            w3=w3,
            contract_abi=abi,
            address=address,
            decode_tuples=decode_tuples,
            function_identifier="enableIrm",
        )
        self.enableLltv = IMorphoEnableLltvContractFunction.factory(
            "enableLltv",
            w3=w3,
            contract_abi=abi,
            address=address,
            decode_tuples=decode_tuples,
            function_identifier="enableLltv",
        )
        self.extSloads = IMorphoExtSloadsContractFunction.factory(
            "extSloads",
            w3=w3,
            contract_abi=abi,
            address=address,
            decode_tuples=decode_tuples,
            function_identifier="extSloads",
        )
        self.feeRecipient = IMorphoFeeRecipientContractFunction.factory(
            "feeRecipient",
            w3=w3,
            contract_abi=abi,
            address=address,
            decode_tuples=decode_tuples,
            function_identifier="feeRecipient",
        )
        self.flashLoan = IMorphoFlashLoanContractFunction.factory(
            "flashLoan",
            w3=w3,
            contract_abi=abi,
            address=address,
            decode_tuples=decode_tuples,
            function_identifier="flashLoan",
        )
        self.idToMarketParams = IMorphoIdToMarketParamsContractFunction.factory(
            "idToMarketParams",
            w3=w3,
            contract_abi=abi,
            address=address,
            decode_tuples=decode_tuples,
            function_identifier="idToMarketParams",
        )
        self.isAuthorized = IMorphoIsAuthorizedContractFunction.factory(
            "isAuthorized",
            w3=w3,
            contract_abi=abi,
            address=address,
            decode_tuples=decode_tuples,
            function_identifier="isAuthorized",
        )
        self.isIrmEnabled = IMorphoIsIrmEnabledContractFunction.factory(
            "isIrmEnabled",
            w3=w3,
            contract_abi=abi,
            address=address,
            decode_tuples=decode_tuples,
            function_identifier="isIrmEnabled",
        )
        self.isLltvEnabled = IMorphoIsLltvEnabledContractFunction.factory(
            "isLltvEnabled",
            w3=w3,
            contract_abi=abi,
            address=address,
            decode_tuples=decode_tuples,
            function_identifier="isLltvEnabled",
        )
        self.liquidate = IMorphoLiquidateContractFunction.factory(
            "liquidate",
            w3=w3,
            contract_abi=abi,
            address=address,
            decode_tuples=decode_tuples,
            function_identifier="liquidate",
        )
        self.market = IMorphoMarketContractFunction.factory(
            "market",
            w3=w3,
            contract_abi=abi,
            address=address,
            decode_tuples=decode_tuples,
            function_identifier="market",
        )
        self.nonce = IMorphoNonceContractFunction.factory(
            "nonce",
            w3=w3,
            contract_abi=abi,
            address=address,
            decode_tuples=decode_tuples,
            function_identifier="nonce",
        )
        self.owner = IMorphoOwnerContractFunction.factory(
            "owner",
            w3=w3,
            contract_abi=abi,
            address=address,
            decode_tuples=decode_tuples,
            function_identifier="owner",
        )
        self.position = IMorphoPositionContractFunction.factory(
            "position",
            w3=w3,
            contract_abi=abi,
            address=address,
            decode_tuples=decode_tuples,
            function_identifier="position",
        )
        self.repay = IMorphoRepayContractFunction.factory(
            "repay",
            w3=w3,
            contract_abi=abi,
            address=address,
            decode_tuples=decode_tuples,
            function_identifier="repay",
        )
        self.setAuthorization = IMorphoSetAuthorizationContractFunction.factory(
            "setAuthorization",
            w3=w3,
            contract_abi=abi,
            address=address,
            decode_tuples=decode_tuples,
            function_identifier="setAuthorization",
        )
        self.setAuthorizationWithSig = (
            IMorphoSetAuthorizationWithSigContractFunction.factory(
                "setAuthorizationWithSig",
                w3=w3,
                contract_abi=abi,
                address=address,
                decode_tuples=decode_tuples,
                function_identifier="setAuthorizationWithSig",
            )
        )
        self.setFee = IMorphoSetFeeContractFunction.factory(
            "setFee",
            w3=w3,
            contract_abi=abi,
            address=address,
            decode_tuples=decode_tuples,
            function_identifier="setFee",
        )
        self.setFeeRecipient = IMorphoSetFeeRecipientContractFunction.factory(
            "setFeeRecipient",
            w3=w3,
            contract_abi=abi,
            address=address,
            decode_tuples=decode_tuples,
            function_identifier="setFeeRecipient",
        )
        self.setOwner = IMorphoSetOwnerContractFunction.factory(
            "setOwner",
            w3=w3,
            contract_abi=abi,
            address=address,
            decode_tuples=decode_tuples,
            function_identifier="setOwner",
        )
        self.supply = IMorphoSupplyContractFunction.factory(
            "supply",
            w3=w3,
            contract_abi=abi,
            address=address,
            decode_tuples=decode_tuples,
            function_identifier="supply",
        )
        self.supplyCollateral = IMorphoSupplyCollateralContractFunction.factory(
            "supplyCollateral",
            w3=w3,
            contract_abi=abi,
            address=address,
            decode_tuples=decode_tuples,
            function_identifier="supplyCollateral",
        )
        self.withdraw = IMorphoWithdrawContractFunction.factory(
            "withdraw",
            w3=w3,
            contract_abi=abi,
            address=address,
            decode_tuples=decode_tuples,
            function_identifier="withdraw",
        )
        self.withdrawCollateral = (
            IMorphoWithdrawCollateralContractFunction.factory(
                "withdrawCollateral",
                w3=w3,
                contract_abi=abi,
                address=address,
                decode_tuples=decode_tuples,
                function_identifier="withdrawCollateral",
            )
        )


imorpho_abi: ABI = cast(
    ABI,
    [
        {
            "type": "function",
            "name": "DOMAIN_SEPARATOR",
            "inputs": [],
            "outputs": [
                {"name": "", "type": "bytes32", "internalType": "bytes32"}
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "accrueInterest",
            "inputs": [
                {
                    "name": "marketParams",
                    "type": "tuple",
                    "internalType": "struct MarketParams",
                    "components": [
                        {
                            "name": "loanToken",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "collateralToken",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "oracle",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "irm",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "lltv",
                            "type": "uint256",
                            "internalType": "uint256",
                        },
                    ],
                }
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "borrow",
            "inputs": [
                {
                    "name": "marketParams",
                    "type": "tuple",
                    "internalType": "struct MarketParams",
                    "components": [
                        {
                            "name": "loanToken",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "collateralToken",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "oracle",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "irm",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "lltv",
                            "type": "uint256",
                            "internalType": "uint256",
                        },
                    ],
                },
                {
                    "name": "assets",
                    "type": "uint256",
                    "internalType": "uint256",
                },
                {
                    "name": "shares",
                    "type": "uint256",
                    "internalType": "uint256",
                },
                {
                    "name": "onBehalf",
                    "type": "address",
                    "internalType": "address",
                },
                {
                    "name": "receiver",
                    "type": "address",
                    "internalType": "address",
                },
            ],
            "outputs": [
                {
                    "name": "assetsBorrowed",
                    "type": "uint256",
                    "internalType": "uint256",
                },
                {
                    "name": "sharesBorrowed",
                    "type": "uint256",
                    "internalType": "uint256",
                },
            ],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "createMarket",
            "inputs": [
                {
                    "name": "marketParams",
                    "type": "tuple",
                    "internalType": "struct MarketParams",
                    "components": [
                        {
                            "name": "loanToken",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "collateralToken",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "oracle",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "irm",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "lltv",
                            "type": "uint256",
                            "internalType": "uint256",
                        },
                    ],
                }
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "enableIrm",
            "inputs": [
                {"name": "irm", "type": "address", "internalType": "address"}
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "enableLltv",
            "inputs": [
                {"name": "lltv", "type": "uint256", "internalType": "uint256"}
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "extSloads",
            "inputs": [
                {
                    "name": "slots",
                    "type": "bytes32[]",
                    "internalType": "bytes32[]",
                }
            ],
            "outputs": [
                {"name": "", "type": "bytes32[]", "internalType": "bytes32[]"}
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "feeRecipient",
            "inputs": [],
            "outputs": [
                {"name": "", "type": "address", "internalType": "address"}
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "flashLoan",
            "inputs": [
                {"name": "token", "type": "address", "internalType": "address"},
                {
                    "name": "assets",
                    "type": "uint256",
                    "internalType": "uint256",
                },
                {"name": "data", "type": "bytes", "internalType": "bytes"},
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "idToMarketParams",
            "inputs": [{"name": "id", "type": "bytes32", "internalType": "Id"}],
            "outputs": [
                {
                    "name": "",
                    "type": "tuple",
                    "internalType": "struct MarketParams",
                    "components": [
                        {
                            "name": "loanToken",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "collateralToken",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "oracle",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "irm",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "lltv",
                            "type": "uint256",
                            "internalType": "uint256",
                        },
                    ],
                }
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "isAuthorized",
            "inputs": [
                {
                    "name": "authorizer",
                    "type": "address",
                    "internalType": "address",
                },
                {
                    "name": "authorized",
                    "type": "address",
                    "internalType": "address",
                },
            ],
            "outputs": [{"name": "", "type": "bool", "internalType": "bool"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "isIrmEnabled",
            "inputs": [
                {"name": "irm", "type": "address", "internalType": "address"}
            ],
            "outputs": [{"name": "", "type": "bool", "internalType": "bool"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "isLltvEnabled",
            "inputs": [
                {"name": "lltv", "type": "uint256", "internalType": "uint256"}
            ],
            "outputs": [{"name": "", "type": "bool", "internalType": "bool"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "liquidate",
            "inputs": [
                {
                    "name": "marketParams",
                    "type": "tuple",
                    "internalType": "struct MarketParams",
                    "components": [
                        {
                            "name": "loanToken",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "collateralToken",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "oracle",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "irm",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "lltv",
                            "type": "uint256",
                            "internalType": "uint256",
                        },
                    ],
                },
                {
                    "name": "borrower",
                    "type": "address",
                    "internalType": "address",
                },
                {
                    "name": "seizedAssets",
                    "type": "uint256",
                    "internalType": "uint256",
                },
                {
                    "name": "repaidShares",
                    "type": "uint256",
                    "internalType": "uint256",
                },
                {"name": "data", "type": "bytes", "internalType": "bytes"},
            ],
            "outputs": [
                {"name": "", "type": "uint256", "internalType": "uint256"},
                {"name": "", "type": "uint256", "internalType": "uint256"},
            ],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "market",
            "inputs": [{"name": "id", "type": "bytes32", "internalType": "Id"}],
            "outputs": [
                {
                    "name": "m",
                    "type": "tuple",
                    "internalType": "struct Market",
                    "components": [
                        {
                            "name": "totalSupplyAssets",
                            "type": "uint128",
                            "internalType": "uint128",
                        },
                        {
                            "name": "totalSupplyShares",
                            "type": "uint128",
                            "internalType": "uint128",
                        },
                        {
                            "name": "totalBorrowAssets",
                            "type": "uint128",
                            "internalType": "uint128",
                        },
                        {
                            "name": "totalBorrowShares",
                            "type": "uint128",
                            "internalType": "uint128",
                        },
                        {
                            "name": "lastUpdate",
                            "type": "uint128",
                            "internalType": "uint128",
                        },
                        {
                            "name": "fee",
                            "type": "uint128",
                            "internalType": "uint128",
                        },
                    ],
                }
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "nonce",
            "inputs": [
                {
                    "name": "authorizer",
                    "type": "address",
                    "internalType": "address",
                }
            ],
            "outputs": [
                {"name": "", "type": "uint256", "internalType": "uint256"}
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "owner",
            "inputs": [],
            "outputs": [
                {"name": "", "type": "address", "internalType": "address"}
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "position",
            "inputs": [
                {"name": "id", "type": "bytes32", "internalType": "Id"},
                {"name": "user", "type": "address", "internalType": "address"},
            ],
            "outputs": [
                {
                    "name": "p",
                    "type": "tuple",
                    "internalType": "struct Position",
                    "components": [
                        {
                            "name": "supplyShares",
                            "type": "uint256",
                            "internalType": "uint256",
                        },
                        {
                            "name": "borrowShares",
                            "type": "uint128",
                            "internalType": "uint128",
                        },
                        {
                            "name": "collateral",
                            "type": "uint128",
                            "internalType": "uint128",
                        },
                    ],
                }
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "repay",
            "inputs": [
                {
                    "name": "marketParams",
                    "type": "tuple",
                    "internalType": "struct MarketParams",
                    "components": [
                        {
                            "name": "loanToken",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "collateralToken",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "oracle",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "irm",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "lltv",
                            "type": "uint256",
                            "internalType": "uint256",
                        },
                    ],
                },
                {
                    "name": "assets",
                    "type": "uint256",
                    "internalType": "uint256",
                },
                {
                    "name": "shares",
                    "type": "uint256",
                    "internalType": "uint256",
                },
                {
                    "name": "onBehalf",
                    "type": "address",
                    "internalType": "address",
                },
                {"name": "data", "type": "bytes", "internalType": "bytes"},
            ],
            "outputs": [
                {
                    "name": "assetsRepaid",
                    "type": "uint256",
                    "internalType": "uint256",
                },
                {
                    "name": "sharesRepaid",
                    "type": "uint256",
                    "internalType": "uint256",
                },
            ],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "setAuthorization",
            "inputs": [
                {
                    "name": "authorized",
                    "type": "address",
                    "internalType": "address",
                },
                {
                    "name": "newIsAuthorized",
                    "type": "bool",
                    "internalType": "bool",
                },
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "setAuthorizationWithSig",
            "inputs": [
                {
                    "name": "authorization",
                    "type": "tuple",
                    "internalType": "struct Authorization",
                    "components": [
                        {
                            "name": "authorizer",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "authorized",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "isAuthorized",
                            "type": "bool",
                            "internalType": "bool",
                        },
                        {
                            "name": "nonce",
                            "type": "uint256",
                            "internalType": "uint256",
                        },
                        {
                            "name": "deadline",
                            "type": "uint256",
                            "internalType": "uint256",
                        },
                    ],
                },
                {
                    "name": "signature",
                    "type": "tuple",
                    "internalType": "struct Signature",
                    "components": [
                        {"name": "v", "type": "uint8", "internalType": "uint8"},
                        {
                            "name": "r",
                            "type": "bytes32",
                            "internalType": "bytes32",
                        },
                        {
                            "name": "s",
                            "type": "bytes32",
                            "internalType": "bytes32",
                        },
                    ],
                },
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "setFee",
            "inputs": [
                {
                    "name": "marketParams",
                    "type": "tuple",
                    "internalType": "struct MarketParams",
                    "components": [
                        {
                            "name": "loanToken",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "collateralToken",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "oracle",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "irm",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "lltv",
                            "type": "uint256",
                            "internalType": "uint256",
                        },
                    ],
                },
                {
                    "name": "newFee",
                    "type": "uint256",
                    "internalType": "uint256",
                },
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "setFeeRecipient",
            "inputs": [
                {
                    "name": "newFeeRecipient",
                    "type": "address",
                    "internalType": "address",
                }
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "setOwner",
            "inputs": [
                {
                    "name": "newOwner",
                    "type": "address",
                    "internalType": "address",
                }
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "supply",
            "inputs": [
                {
                    "name": "marketParams",
                    "type": "tuple",
                    "internalType": "struct MarketParams",
                    "components": [
                        {
                            "name": "loanToken",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "collateralToken",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "oracle",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "irm",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "lltv",
                            "type": "uint256",
                            "internalType": "uint256",
                        },
                    ],
                },
                {
                    "name": "assets",
                    "type": "uint256",
                    "internalType": "uint256",
                },
                {
                    "name": "shares",
                    "type": "uint256",
                    "internalType": "uint256",
                },
                {
                    "name": "onBehalf",
                    "type": "address",
                    "internalType": "address",
                },
                {"name": "data", "type": "bytes", "internalType": "bytes"},
            ],
            "outputs": [
                {
                    "name": "assetsSupplied",
                    "type": "uint256",
                    "internalType": "uint256",
                },
                {
                    "name": "sharesSupplied",
                    "type": "uint256",
                    "internalType": "uint256",
                },
            ],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "supplyCollateral",
            "inputs": [
                {
                    "name": "marketParams",
                    "type": "tuple",
                    "internalType": "struct MarketParams",
                    "components": [
                        {
                            "name": "loanToken",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "collateralToken",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "oracle",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "irm",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "lltv",
                            "type": "uint256",
                            "internalType": "uint256",
                        },
                    ],
                },
                {
                    "name": "assets",
                    "type": "uint256",
                    "internalType": "uint256",
                },
                {
                    "name": "onBehalf",
                    "type": "address",
                    "internalType": "address",
                },
                {"name": "data", "type": "bytes", "internalType": "bytes"},
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "withdraw",
            "inputs": [
                {
                    "name": "marketParams",
                    "type": "tuple",
                    "internalType": "struct MarketParams",
                    "components": [
                        {
                            "name": "loanToken",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "collateralToken",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "oracle",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "irm",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "lltv",
                            "type": "uint256",
                            "internalType": "uint256",
                        },
                    ],
                },
                {
                    "name": "assets",
                    "type": "uint256",
                    "internalType": "uint256",
                },
                {
                    "name": "shares",
                    "type": "uint256",
                    "internalType": "uint256",
                },
                {
                    "name": "onBehalf",
                    "type": "address",
                    "internalType": "address",
                },
                {
                    "name": "receiver",
                    "type": "address",
                    "internalType": "address",
                },
            ],
            "outputs": [
                {
                    "name": "assetsWithdrawn",
                    "type": "uint256",
                    "internalType": "uint256",
                },
                {
                    "name": "sharesWithdrawn",
                    "type": "uint256",
                    "internalType": "uint256",
                },
            ],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "withdrawCollateral",
            "inputs": [
                {
                    "name": "marketParams",
                    "type": "tuple",
                    "internalType": "struct MarketParams",
                    "components": [
                        {
                            "name": "loanToken",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "collateralToken",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "oracle",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "irm",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "lltv",
                            "type": "uint256",
                            "internalType": "uint256",
                        },
                    ],
                },
                {
                    "name": "assets",
                    "type": "uint256",
                    "internalType": "uint256",
                },
                {
                    "name": "onBehalf",
                    "type": "address",
                    "internalType": "address",
                },
                {
                    "name": "receiver",
                    "type": "address",
                    "internalType": "address",
                },
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
    ],
)
# pylint: disable=line-too-long
imorpho_bytecode = HexStr("0x")


class IMorphoContract(Contract):
    """A web3.py Contract class for the IMorpho contract."""

    abi: ABI = imorpho_abi
    bytecode: bytes | None = try_bytecode_hexbytes(imorpho_bytecode, "imorpho")

    def __init__(self, address: ChecksumAddress | None = None) -> None:
        try:
            # Initialize parent Contract class
            super().__init__(address=address)
            self.functions = IMorphoContractFunctions(imorpho_abi, self.w3, address)  # type: ignore

        except FallbackNotFound:
            print("Fallback function not found. Continuing...")

    functions: IMorphoContractFunctions

    @classmethod
    def constructor(cls) -> ContractConstructor:  # type: ignore
        """Creates a transaction with the contract's constructor function.

        Parameters
        ----------

        w3 : Web3
            A web3 instance.
        account : LocalAccount
            The account to use to deploy the contract.

        Returns
        -------
        Self
            A deployed instance of the contract.

        """

        return super().constructor()

    @classmethod
    def deploy(cls, w3: Web3, account: LocalAccount | ChecksumAddress) -> Self:
        """Deploys and instance of the contract.

        Parameters
        ----------
        w3 : Web3
            A web3 instance.
        account : LocalAccount
            The account to use to deploy the contract.

        Returns
        -------
        Self
            A deployed instance of the contract.
        """
        deployer = cls.factory(w3=w3)
        constructor_fn = deployer.constructor()

        # if an address is supplied, try to use a web3 default account
        if isinstance(account, str):
            tx_hash = constructor_fn.transact({"from": account})
            tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash)

            deployed_contract = deployer(address=tx_receipt.contractAddress)  # type: ignore
            return deployed_contract

        # otherwise use the account provided.
        deployment_tx = constructor_fn.build_transaction()
        current_nonce = w3.eth.get_transaction_count(account.address, "pending")
        deployment_tx.update({"nonce": current_nonce})

        # Sign the transaction with local account private key
        signed_tx = account.sign_transaction(deployment_tx)

        # Send the signed transaction and wait for receipt
        tx_hash = w3.eth.send_raw_transaction(signed_tx.rawTransaction)
        tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash)

        deployed_contract = deployer(address=tx_receipt.contractAddress)  # type: ignore
        return deployed_contract

    @classmethod
    def factory(
        cls, w3: Web3, class_name: str | None = None, **kwargs: Any
    ) -> Type[Self]:
        """Deploys and instance of the contract.

        Parameters
        ----------
        w3 : Web3
            A web3 instance.
        class_name: str | None
            The instance class name.

        Returns
        -------
        Self
            A deployed instance of the contract.
        """
        contract = super().factory(w3, class_name, **kwargs)
        contract.functions = IMorphoContractFunctions(imorpho_abi, w3, None)

        return contract
